==> core-0.3.1-sources.jar: main/gay/zharel/hermes/math/MathUtil.kt <==
@file:JvmName("Math")

package gay.zharel.hermes.math

import kotlinx.serialization.Serializable

// ~10 * machine epsilon
private const val EPS = 2.2e-15

/**
 * Returns a small value with the same sign as [x].
 *
 * Function \(snz(x)\) from section VI.A of the [SymForce paper](https://arxiv.org/abs/2204.07889) for use in
 * singularity handling.
 * Returns a small positive value if [x] is non-negative,
 * otherwise a small negative value.
 *
 * @param x The input value.
 * @return A small value with the same sign as [x].
 */
fun snz(x: Double) =
    if (x >= 0.0) {
        EPS
    } else {
        -EPS
    }

/**
 * Clamps [x] to the range [[lo], [hi]].
 *
 * @param x The value to clamp.
 * @param lo The lower bound.
 * @param hi The upper bound.
 * @return [x] if it is within the range, otherwise [lo] or [hi].
 */
fun clamp(x: Double, lo: Double, hi: Double): Double {
    if (x < lo) {
        return lo
    }
    if (x > hi) {
        return hi
    }
    return x
}

/**
 * Data class representing a minimum and maximum value pair.
 *
 * @property min The minimum value.
 * @property max The maximum value.
 */
data class MinMax(@JvmField val min: Double, @JvmField val max: Double)

/**
 * @usesMathJax
 *
 * Partitions \([a, b]\) into \((n - 1)\) equal intervals and returns the endpoints.
 *
 * @param begin \(a\) The start of the range.
 * @param end \(b\) The end of the range.
 * @param samples \(n\) The number of samples (must be at least 2).
 * @return List of endpoints of the intervals.
 * @throws IllegalArgumentException if [samples] < 2.
 */
fun range(begin: Double, end: Double, samples: Int): List<Double> {
    require(samples >= 2) { "Number of samples ($samples) must be at least 2" }
    val dx = (end - begin) / (samples - 1)
    return (0 until samples).map { begin + dx * it }
}

/**
 * @usesMathJax
 *
 * Partitions \([a, b]\) into \(n\) equal intervals and returns the center values.
 *
 * @param begin \(a\) The start of the range.
 * @param end \(b\) The end of the range.
 * @param samples \(n\) The number of samples (must be at least 1).
 * @return List of center values of the intervals.
 * @throws IllegalArgumentException if [samples] < 1.
 */
fun rangeCentered(begin: Double, end: Double, samples: Int): List<Double> {
    require(samples >= 1) { "Number of samples must be at least 1" }
    val dx = (end - begin) / samples
    return (0 until samples).map { begin + 0.5 * dx + dx * it }
}

/**
 * Data class representing the result of an integral scan.
 *
 * @property values The sample points.
 * @property sums The cumulative integrals at each sample point.
 */
@Serializable
data class IntegralScanResult(
    @JvmField
    val values: List<Double>,
    @JvmField
    val sums: List<Double>,
)

/**
 * @usesMathJax
 *
 * Returns samples of \(g(t) = \int_a^t f(x) \, dx\) for various values \(a \leq t \leq b\). The sampling points are
 * chosen adaptively using the algorithm `adaptsim` from [Gander and Gautschi](https://doi.org/10.1023/A:1022318402393).
 *
 * @param a The lower bound of integration.
 * @param b The upper bound of integration.
 * @param eps Desired error in the length approximation \(g(b)\).
 * @param f The function to integrate.
 * @return [IntegralScanResult] containing sample points and their cumulative integrals.
 */
fun integralScan(a: Double, b: Double, eps: Double, f: (Double) -> Double): IntegralScanResult {
    val m = (a + b) / 2
    val fa = f(a)
    val fm = f(m)
    val fb = f(b)

    var i = (b - a) / 8 * (
        fa + fm + fb +
            f(a + 0.9501 * (b - a)) +
            f(a + 0.2311 * (b - a)) +
            f(a + 0.6068 * (b - a)) +
            f(a + 0.4860 * (b - a)) +
            f(a + 0.8913 * (b - a))
        )
    if (i == 0.0) {
        i = b - a
    }
    i *= eps / Math.ulp(1.0)

    val values = mutableListOf(0.0)
    val sums = mutableListOf(0.0)

    fun helper(a: Double, m: Double, b: Double, fa: Double, fm: Double, fb: Double) {
        val h = (b - a) / 4
        val ml = a + h
        val mr = b - h
        val fml = f(ml)
        val fmr = f(mr)
        var i1 = h / 1.5 * (fa + 4 * fm + fb)
        val i2 = h / 3 * (fa + 4 * (fml + fmr) + 2 * fm + fb)
        i1 = (16 * i2 - i1) / 15
        if (i + (i1 - i2) == i || m <= a || b <= m) {
            values.add(b)
            sums.add(sums.last() + i1)
        } else {
            helper(a, ml, m, fa, fml, fm)
            helper(m, mr, b, fm, fmr, fb)
        }
    }

    helper(a, m, b, fa, fm, fb)

    return IntegralScanResult(values, sums)
}

/**
 * Computes the factorial of this integer.
 *
 * @receiver The integer to compute the factorial of (must be non-negative).
 * @return The factorial of this integer.
 * @throws IllegalArgumentException if the integer is negative.
 */
fun Int.fact(): Int {
    require(this >= 0)
    return when (this) {
        0 -> 1
        1 -> 1
        else -> this * (this-1).fact()
    }
}

==> core-0.3.4-sources.jar: main/gay/zharel/hermes/math/MathUtil.kt <==
/*
 * Copyright (c) 2025 Hermes FTC
 *
 * Use of this source code is governed by an MIT-style
 * license that can be found in the LICENSE file at the root of this repository or at
 * https://opensource.org/licenses/MIT.
 */

@file:JvmName("Math")

package gay.zharel.hermes.math

import kotlinx.serialization.Serializable
import kotlin.math.sin

// ~10 * machine epsilon
private const val EPS = 2.2e-15

/**
 * Returns a small value with the same sign as [x].
 *
 * Function \(snz(x)\) from section VI.A of the [SymForce paper](https://arxiv.org/abs/2204.07889) for use in
 * singularity handling.
 * Returns a small positive value if [x] is non-negative,
 * otherwise a small negative value.
 *
 * @param x The input value.
 * @return A small value with the same sign as [x].
 */
fun snz(x: Double) =
    if (x >= 0.0) {
        EPS
    } else {
        -EPS
    }

/**
 * @usesMathJax
 *
 * Returns the sine of \(x + snz(x)\).
 */
fun sinc(x: Double): Double {
    val u = x + snz(x)
    return sin(u) / u
}

/**
 * Clamps [x] to the range [[lo], [hi]].
 *
 * @param x The value to clamp.
 * @param lo The lower bound.
 * @param hi The upper bound.
 * @return [x] if it is within the range, otherwise [lo] or [hi].
 */
fun clamp(x: Double, lo: Double, hi: Double): Double {
    if (x < lo) {
        return lo
    }
    if (x > hi) {
        return hi
    }
    return x
}

/**
 * Data class representing a minimum and maximum value pair.
 *
 * @property min The minimum value.
 * @property max The maximum value.
 */
data class MinMax(@JvmField val min: Double, @JvmField val max: Double)

/**
 * @usesMathJax
 *
 * Partitions \([a, b]\) into \((n - 1)\) equal intervals and returns the endpoints.
 *
 * @param begin \(a\) The start of the range.
 * @param end \(b\) The end of the range.
 * @param samples \(n\) The number of samples (must be at least 2).
 * @return List of endpoints of the intervals.
 * @throws IllegalArgumentException if [samples] < 2.
 */
fun range(begin: Double, end: Double, samples: Int): List<Double> {
    require(samples >= 2) { "Number of samples ($samples) must be at least 2" }
    val dx = (end - begin) / (samples - 1)
    return (0 until samples).map { begin + dx * it }
}

/**
 * @usesMathJax
 *
 * Partitions \([a, b]\) into \(n\) equal intervals and returns the center values.
 *
 * @param begin \(a\) The start of the range.
 * @param end \(b\) The end of the range.
 * @param samples \(n\) The number of samples (must be at least 1).
 * @return List of center values of the intervals.
 * @throws IllegalArgumentException if [samples] < 1.
 */
fun rangeCentered(begin: Double, end: Double, samples: Int): List<Double> {
    require(samples >= 1) { "Number of samples must be at least 1" }
    val dx = (end - begin) / samples
    return (0 until samples).map { begin + 0.5 * dx + dx * it }
}

/**
 * Data class representing the result of an integral scan.
 *
 * @property values The sample points.
 * @property sums The cumulative integrals at each sample point.
 */
@Serializable
data class IntegralScanResult(
    @JvmField
    val values: List<Double>,
    @JvmField
    val sums: List<Double>,
)

/**
 * @usesMathJax
 *
 * Returns samples of \(g(t) = \int_a^t f(x) \, dx\) for various values \(a \leq t \leq b\). The sampling points are
 * chosen adaptively using the algorithm `adaptsim` from [Gander and Gautschi](https://doi.org/10.1023/A:1022318402393).
 *
 * @param a The lower bound of integration.
 * @param b The upper bound of integration.
 * @param eps Desired error in the length approximation \(g(b)\).
 * @param f The function to integrate.
 * @return [IntegralScanResult] containing sample points and their cumulative integrals.
 */
fun integralScan(a: Double, b: Double, eps: Double, f: (Double) -> Double): IntegralScanResult {
    val m = (a + b) / 2
    val fa = f(a)
    val fm = f(m)
    val fb = f(b)

    var i = (b - a) / 8 * (
        fa + fm + fb +
            f(a + 0.9501 * (b - a)) +
            f(a + 0.2311 * (b - a)) +
            f(a + 0.6068 * (b - a)) +
            f(a + 0.4860 * (b - a)) +
            f(a + 0.8913 * (b - a))
        )
    if (i == 0.0) {
        i = b - a
    }
    i *= eps / Math.ulp(1.0)

    val values = mutableListOf(0.0)
    val sums = mutableListOf(0.0)

    fun helper(a: Double, m: Double, b: Double, fa: Double, fm: Double, fb: Double) {
        val h = (b - a) / 4
        val ml = a + h
        val mr = b - h
        val fml = f(ml)
        val fmr = f(mr)
        var i1 = h / 1.5 * (fa + 4 * fm + fb)
        val i2 = h / 3 * (fa + 4 * (fml + fmr) + 2 * fm + fb)
        i1 = (16 * i2 - i1) / 15
        if (i + (i1 - i2) == i || m <= a || b <= m) {
            values.add(b)
            sums.add(sums.last() + i1)
        } else {
            helper(a, ml, m, fa, fml, fm)
            helper(m, mr, b, fm, fmr, fb)
        }
    }

    helper(a, m, b, fa, fm, fb)

    return IntegralScanResult(values, sums)
}

/**
 * Computes the factorial of this integer.
 *
 * @receiver The integer to compute the factorial of (must be non-negative).
 * @return The factorial of this integer.
 * @throws IllegalArgumentException if the integer is negative.
 */
fun Int.fact(): Int {
    require(this >= 0)
    return when (this) {
        0 -> 1
        1 -> 1
        else -> this * (this-1).fact()
    }
}
